@@ -1,7 +1,8 @@
+use super::{code::CodeBlockParseError, elements::SourcePosition};
 use crate::{
-    markdown::elements::{
-        Code, ListItem, ListItemType, MarkdownElement, ParagraphElement, ProgrammingLanguage, StyledText, Table,
-        TableRow, Text,
+    markdown::{
+        code::CodeBlockParser,
+        elements::{ListItem, ListItemType, MarkdownElement, ParagraphElement, StyledText, Table, TableRow, Text},
     },
     style::TextStyle,
 };
@@ -19,7 +20,7 @@ use std::{
 };
 
 /// The result of parsing a markdown file.
-pub type ParseResult<T> = Result<T, ParseError>;
+pub(crate) type ParseResult<T> = Result<T, ParseError>;
 
 struct ParserOptions(ComrakOptions);
 
@@ -48,16 +49,42 @@ impl<'a> MarkdownParser<'a> {
     }
 
     /// Parse the contents of a markdown file.
-    pub fn parse(&self, contents: &str) -> ParseResult<Vec<MarkdownElement>> {
+    pub(crate) fn parse(&self, contents: &str) -> ParseResult<Vec<MarkdownElement>> {
         let node = parse_document(self.arena, contents, &self.options);
         let mut elements = Vec::new();
+        let mut lines_offset = 0;
         for node in node.children() {
-            let element = Self::parse_node(node)?;
-            elements.extend(element);
+            let mut parsed_elements =
+                Self::parse_node(node).map_err(|e| ParseError::new(e.kind, e.sourcepos.offset_lines(lines_offset)))?;
+            if let Some(MarkdownElement::FrontMatter(contents)) = parsed_elements.first() {
+                lines_offset += contents.lines().count() + 2;
+            }
+            // comrak ignores the lines in the front matter so we need to offset this ourselves.
+            Self::adjust_source_positions(parsed_elements.iter_mut(), lines_offset);
+            elements.extend(parsed_elements);
         }
         Ok(elements)
     }
 
+    fn adjust_source_positions<'b>(elements: impl Iterator<Item = &'b mut MarkdownElement>, lines_offset: usize) {
+        for element in elements {
+            let position = match element {
+                MarkdownElement::FrontMatter(_)
+                | MarkdownElement::SetexHeading { .. }
+                | MarkdownElement::Heading { .. }
+                | MarkdownElement::Paragraph(_)
+                | MarkdownElement::Image { .. }
+                | MarkdownElement::List(_)
+                | MarkdownElement::Code(_)
+                | MarkdownElement::Table(_)
+                | MarkdownElement::ThematicBreak
+                | MarkdownElement::BlockQuote(_) => continue,
+                MarkdownElement::Comment { source_position, .. } => source_position,
+            };
+            *position = position.offset_lines(lines_offset);
+        }
+    }
+
     fn parse_node(node: &'a AstNode<'a>) -> ParseResult<Vec<MarkdownElement>> {
         let data = node.data.borrow();
         let element = match &data.value {
@@ -97,7 +124,7 @@ impl<'a> MarkdownParser<'a> {
         }
         let block = &block[start_tag.len()..];
         let block = &block[0..block.len() - end_tag.len()];
-        Ok(MarkdownElement::Comment(block.into()))
+        Ok(MarkdownElement::Comment { comment: block.into(), source_position: sourcepos.into() })
     }
 
     fn parse_block_quote(node: &'a AstNode<'a>) -> ParseResult<MarkdownElement> {
@@ -123,42 +150,8 @@ impl<'a> MarkdownParser<'a> {
         if !block.fenced {
             return Err(ParseErrorKind::UnfencedCodeBlock.with_sourcepos(sourcepos));
         }
-        use ProgrammingLanguage::*;
-        let language = match block.info.as_str() {
-            "asp" => Asp,
-            "bash" => Bash,
-            "c" => C,
-            "csharp" => CSharp,
-            "clojure" => Clojure,
-            "cpp" | "c++" => Cpp,
-            "css" => Css,
-            "d" => DLang,
-            "erlang" => Erlang,
-            "go" => Go,
-            "haskell" => Haskell,
-            "html" => Html,
-            "java" => Java,
-            "javascript" | "js" => JavaScript,
-            "json" => Json,
-            "latex" => Latex,
-            "lua" => Lua,
-            "make" => Makefile,
-            "markdown" => Markdown,
-            "ocaml" => OCaml,
-            "perl" => Perl,
-            "php" => Php,
-            "python" => Python,
-            "r" => R,
-            "rust" => Rust,
-            "scala" => Scala,
-            "shell" | "sh" | "zsh" | "fish" => Shell,
-            "sql" => Sql,
-            "typescript" | "ts" => TypeScript,
-            "xml" => Xml,
-            "yaml" => Yaml,
-            _ => Unknown,
-        };
-        let code = Code { contents: block.literal.clone(), language };
+        let code =
+            CodeBlockParser::parse(block).map_err(|e| ParseErrorKind::InvalidCodeBlock(e).with_sourcepos(sourcepos))?;
         Ok(MarkdownElement::Code(code))
     }
 
@@ -183,7 +176,10 @@ impl<'a> MarkdownParser<'a> {
                     if !paragraph_elements.is_empty() {
                         elements.push(MarkdownElement::Paragraph(mem::take(&mut paragraph_elements)));
                     }
-                    elements.push(MarkdownElement::Image(path.into()));
+                    elements.push(MarkdownElement::Image {
+                        path: path.into(),
+                        source_position: node.data.borrow().sourcepos.into(),
+                    });
                 }
             }
         }
@@ -210,12 +206,11 @@ impl<'a> MarkdownParser<'a> {
 
     fn parse_list(root: &'a AstNode<'a>, depth: u8) -> ParseResult<Vec<ListItem>> {
         let mut elements = Vec::new();
-        for (index, node) in root.children().enumerate() {
-            let number = (index + 1) as u16;
+        for node in root.children() {
             let data = node.data.borrow();
             match &data.value {
                 NodeValue::Item(item) => {
-                    elements.extend(Self::parse_list_item(item, node, depth, number)?);
+                    elements.extend(Self::parse_list_item(item, node, depth)?);
                 }
                 other => {
                     return Err(ParseErrorKind::UnsupportedStructure {
@@ -229,11 +224,11 @@ impl<'a> MarkdownParser<'a> {
         Ok(elements)
     }
 
-    fn parse_list_item(item: &NodeList, root: &'a AstNode<'a>, depth: u8, number: u16) -> ParseResult<Vec<ListItem>> {
+    fn parse_list_item(item: &NodeList, root: &'a AstNode<'a>, depth: u8) -> ParseResult<Vec<ListItem>> {
         let item_type = match (item.list_type, item.delimiter) {
             (ListType::Bullet, _) => ListItemType::Unordered,
-            (ListType::Ordered, ListDelimType::Paren) => ListItemType::OrderedParens(number),
-            (ListType::Ordered, ListDelimType::Period) => ListItemType::OrderedPeriod(number),
+            (ListType::Ordered, ListDelimType::Paren) => ListItemType::OrderedParens,
+            (ListType::Ordered, ListDelimType::Period) => ListItemType::OrderedPeriod,
         };
         let mut elements = Vec::new();
         for node in root.children() {
@@ -378,10 +373,10 @@ impl Inline {
 #[derive(thiserror::Error, Debug)]
 pub struct ParseError {
     /// The kind of error.
-    pub kind: ParseErrorKind,
+    pub(crate) kind: ParseErrorKind,
 
     /// The position in the source file this error originated from.
-    pub sourcepos: Sourcepos,
+    pub(crate) sourcepos: SourcePosition,
 }
 
 impl Display for ParseError {
@@ -391,14 +386,14 @@ impl Display for ParseError {
 }
 
 impl ParseError {
-    fn new(kind: ParseErrorKind, sourcepos: Sourcepos) -> Self {
-        Self { kind, sourcepos }
+    fn new<S: Into<SourcePosition>>(kind: ParseErrorKind, sourcepos: S) -> Self {
+        Self { kind, sourcepos: sourcepos.into() }
     }
 }
 
 /// The kind of error.
 #[derive(Debug)]
-pub enum ParseErrorKind {
+pub(crate) enum ParseErrorKind {
     /// We don't support parsing this element.
     UnsupportedElement(&'static str),
 
@@ -408,6 +403,9 @@ pub enum ParseErrorKind {
     /// We don't support unfenced code blocks.
     UnfencedCodeBlock,
 
+    /// A code block contains invalid attributes.
+    InvalidCodeBlock(CodeBlockParseError),
+
     /// An internal parsing error.
     Internal(String),
 }
@@ -420,13 +418,14 @@ impl Display for ParseErrorKind {
                 write!(f, "unsupported structure in {container}: {element}")
             }
             Self::UnfencedCodeBlock => write!(f, "only fenced code blocks are supported"),
+            Self::InvalidCodeBlock(error) => write!(f, "invalid code block: {error}"),
             Self::Internal(message) => write!(f, "internal error: {message}"),
         }
     }
 }
 
 impl ParseErrorKind {
-    fn with_sourcepos(self, sourcepos: Sourcepos) -> ParseError {
+    fn with_sourcepos<S: Into<SourcePosition>>(self, sourcepos: S) -> ParseError {
         ParseError::new(self, sourcepos)
     }
 }
@@ -475,9 +474,9 @@ impl Identifier for NodeValue {
 
 #[cfg(test)]
 mod test {
-    use std::path::Path;
-
     use super::*;
+    use crate::markdown::elements::CodeLanguage;
+    use std::path::Path;
 
     fn parse_single(input: &str) -> MarkdownElement {
         let arena = Arena::new();
@@ -541,7 +540,7 @@ boop
     #[test]
     fn image() {
         let parsed = parse_single("![](potato.png)");
-        let MarkdownElement::Image(path) = parsed else { panic!("not an image: {parsed:?}") };
+        let MarkdownElement::Image { path, .. } = parsed else { panic!("not an image: {parsed:?}") };
         assert_eq!(path, Path::new("potato.png"));
     }
 
@@ -633,8 +632,23 @@ let q = 42;
 ",
         );
         let MarkdownElement::Code(code) = parsed else { panic!("not a code block: {parsed:?}") };
-        assert_eq!(code.language, ProgrammingLanguage::Rust);
+        assert_eq!(code.language, CodeLanguage::Rust);
         assert_eq!(code.contents, "let q = 42;\n");
+        assert!(!code.attributes.execute);
+    }
+
+    #[test]
+    fn executable_code_block() {
+        let parsed = parse_single(
+            r"
+```bash +exec
+echo hi mom
+````
+",
+        );
+        let MarkdownElement::Code(code) = parsed else { panic!("not a code block: {parsed:?}") };
+        assert_eq!(code.language, CodeLanguage::Shell("bash".into()));
+        assert!(code.attributes.execute);
     }
 
     #[test]
@@ -672,8 +686,8 @@ let q = 42;
 <!-- foo -->
 ",
         );
-        let MarkdownElement::Comment(text) = parsed else { panic!("not a comment: {parsed:?}") };
-        assert_eq!(text, " foo ");
+        let MarkdownElement::Comment { comment, .. } = parsed else { panic!("not a comment: {parsed:?}") };
+        assert_eq!(comment, " foo ");
     }
 
     #[test]
@@ -724,4 +738,38 @@ bye
         assert_eq!(parsed.len(), 3);
         assert!(matches!(parsed[1], MarkdownElement::ThematicBreak));
     }
+
+    #[test]
+    fn error_lines_offset_by_front_matter() {
+        let input = r"---
+hi
+mom
+---
+
+* ![](potato.png)
+";
+        let arena = Arena::new();
+        let result = MarkdownParser::new(&arena).parse(input);
+        let Err(e) = result else {
+            panic!("parsing didn't fail");
+        };
+        assert_eq!(e.sourcepos.start.line, 5);
+        assert_eq!(e.sourcepos.start.column, 3);
+    }
+
+    #[test]
+    fn comment_lines_offset_by_front_matter() {
+        let parsed = parse_all(
+            r"---
+hi
+mom
+---
+
+<!-- hello -->
+",
+        );
+        let MarkdownElement::Comment { source_position, .. } = &parsed[1] else { panic!("not a comment") };
+        assert_eq!(source_position.start.line, 5);
+        assert_eq!(source_position.start.column, 1);
+    }
 }
