@@ -1,10 +1,7 @@
 use crate::{
-    builder::{BuildError, PresentationBuilder},
+    builder::{BuildError, PresentationBuilder, PresentationBuilderOptions},
     diff::PresentationDiffer,
-    input::{
-        source::{Command, CommandSource},
-        user::UserCommand,
-    },
+    input::source::{Command, CommandSource},
     markdown::parse::{MarkdownParser, ParseError},
     presentation::Presentation,
     render::{
@@ -15,6 +12,7 @@ use crate::{
     theme::PresentationTheme,
 };
 use std::{
+    collections::HashSet,
     fs,
     io::{self, Stdout},
     mem,
@@ -32,6 +30,7 @@ pub struct Presenter<'a> {
     resources: Resources,
     mode: PresentMode,
     state: PresenterState,
+    slides_with_pending_widgets: HashSet<usize>,
 }
 
 impl<'a> Presenter<'a> {
@@ -44,7 +43,16 @@ impl<'a> Presenter<'a> {
         resources: Resources,
         mode: PresentMode,
     ) -> Self {
-        Self { default_theme, default_highlighter, commands, parser, resources, mode, state: PresenterState::Empty }
+        Self {
+            default_theme,
+            default_highlighter,
+            commands,
+            parser,
+            resources,
+            mode,
+            state: PresenterState::Empty,
+            slides_with_pending_widgets: HashSet::new(),
+        }
     }
 
     /// Run a presentation.
@@ -54,27 +62,44 @@ impl<'a> Presenter<'a> {
         let mut drawer = TerminalDrawer::new(io::stdout())?;
         loop {
             self.render(&mut drawer)?;
+            self.update_widgets(&mut drawer)?;
 
             loop {
-                let command = match self.commands.next_command()? {
-                    Command::User(command) => command,
-                    Command::ReloadPresentation => {
+                self.update_widgets(&mut drawer)?;
+                let Some(command) = self.commands.try_next_command()? else {
+                    continue;
+                };
+                match self.apply_command(command) {
+                    CommandSideEffect::Exit => return Ok(()),
+                    CommandSideEffect::Reload => {
                         self.try_reload(path);
                         break;
                     }
-                    Command::Abort { error } => return Err(PresentationError::Fatal(error)),
-                };
-                match self.apply_user_command(command) {
-                    CommandSideEffect::Exit => return Ok(()),
                     CommandSideEffect::Redraw => {
                         break;
                     }
+                    CommandSideEffect::PollWidgets => {
+                        self.slides_with_pending_widgets.insert(self.state.presentation().current_slide_index());
+                    }
                     CommandSideEffect::None => (),
                 };
             }
         }
     }
 
+    fn update_widgets(&mut self, drawer: &mut TerminalDrawer<Stdout>) -> RenderResult {
+        let current_index = self.state.presentation().current_slide_index();
+        if self.slides_with_pending_widgets.contains(&current_index) {
+            self.render(drawer)?;
+            if self.state.presentation_mut().widgets_rendered() {
+                // Render one last time just in case it _just_ rendered
+                self.render(drawer)?;
+                self.slides_with_pending_widgets.remove(&current_index);
+            }
+        }
+        Ok(())
+    }
+
     fn render(&mut self, drawer: &mut TerminalDrawer<Stdout>) -> RenderResult {
         let result = match &self.state {
             PresenterState::Presenting(presentation) => drawer.render_slide(presentation),
@@ -86,22 +111,43 @@ impl<'a> Presenter<'a> {
         if matches!(result, Err(RenderError::TerminalTooSmall)) { Ok(()) } else { result }
     }
 
-    fn apply_user_command(&mut self, command: UserCommand) -> CommandSideEffect {
-        // This one always happens no matter our state.
-        if matches!(command, UserCommand::Exit) {
-            return CommandSideEffect::Exit;
-        }
+    fn apply_command(&mut self, command: Command) -> CommandSideEffect {
+        // These ones always happens no matter our state.
+        match command {
+            Command::Reload => {
+                return CommandSideEffect::Reload;
+            }
+            Command::HardReload => {
+                if matches!(self.mode, PresentMode::Development) {
+                    self.resources.clear();
+                }
+                return CommandSideEffect::Reload;
+            }
+            Command::Exit => return CommandSideEffect::Exit,
+            _ => (),
+        };
+
+        // Now apply the commands that require a presentation.
         let PresenterState::Presenting(presentation) = &mut self.state else {
             return CommandSideEffect::None;
         };
         let needs_redraw = match command {
-            UserCommand::Redraw => true,
-            UserCommand::JumpNextSlide => presentation.jump_next_slide(),
-            UserCommand::JumpPreviousSlide => presentation.jump_previous_slide(),
-            UserCommand::JumpFirstSlide => presentation.jump_first_slide(),
-            UserCommand::JumpLastSlide => presentation.jump_last_slide(),
-            UserCommand::JumpSlide(number) => presentation.jump_slide(number.saturating_sub(1) as usize),
-            UserCommand::Exit => return CommandSideEffect::Exit,
+            Command::Redraw => true,
+            Command::JumpNextSlide => presentation.jump_next_slide(),
+            Command::JumpPreviousSlide => presentation.jump_previous_slide(),
+            Command::JumpFirstSlide => presentation.jump_first_slide(),
+            Command::JumpLastSlide => presentation.jump_last_slide(),
+            Command::JumpSlide(number) => presentation.jump_slide(number.saturating_sub(1) as usize),
+            Command::RenderWidgets => {
+                if presentation.render_slide_widgets() {
+                    self.slides_with_pending_widgets.insert(self.state.presentation().current_slide_index());
+                    return CommandSideEffect::PollWidgets;
+                } else {
+                    return CommandSideEffect::None;
+                }
+            }
+            // These are handled above as they don't require the presentation
+            Command::Reload | Command::HardReload | Command::Exit => panic!("unreachable commands"),
         };
         if needs_redraw { CommandSideEffect::Redraw } else { CommandSideEffect::None }
     }
@@ -110,12 +156,17 @@ impl<'a> Presenter<'a> {
         if matches!(self.mode, PresentMode::Presentation) {
             return;
         }
+        self.slides_with_pending_widgets.clear();
         match self.load_presentation(path) {
             Ok(mut presentation) => {
                 let current = self.state.presentation();
-                let target_slide = PresentationDiffer::first_modified_slide(current, &presentation)
-                    .unwrap_or(current.current_slide_index());
-                presentation.jump_slide(target_slide);
+                if let Some(modification) = PresentationDiffer::find_first_modification(current, &presentation) {
+                    presentation.jump_slide(modification.slide_index);
+                    presentation.jump_chunk(modification.chunk_index);
+                } else {
+                    presentation.jump_slide(current.current_slide_index());
+                    presentation.jump_chunk(current.current_chunk());
+                }
                 self.state = PresenterState::Presenting(presentation)
             }
             Err(e) => {
@@ -128,9 +179,17 @@ impl<'a> Presenter<'a> {
     fn load_presentation(&mut self, path: &Path) -> Result<Presentation, LoadPresentationError> {
         let content = fs::read_to_string(path).map_err(LoadPresentationError::Reading)?;
         let elements = self.parser.parse(&content)?;
-        let presentation =
-            PresentationBuilder::new(self.default_highlighter.clone(), self.default_theme, &mut self.resources)
-                .build(elements)?;
+        let mut options = PresentationBuilderOptions::default();
+        if matches!(self.mode, PresentMode::Export) {
+            options.allow_mutations = false;
+        }
+        let presentation = PresentationBuilder::new(
+            self.default_highlighter.clone(),
+            self.default_theme,
+            &mut self.resources,
+            options,
+        )
+        .build(elements)?;
         Ok(presentation)
     }
 }
@@ -138,6 +197,8 @@ impl<'a> Presenter<'a> {
 enum CommandSideEffect {
     Exit,
     Redraw,
+    PollWidgets,
+    Reload,
     None,
 }
 
@@ -161,6 +222,14 @@ impl PresenterState {
         }
     }
 
+    fn presentation_mut(&mut self) -> &mut Presentation {
+        match self {
+            Self::Presenting(presentation) => presentation,
+            Self::Failure { presentation, .. } => presentation,
+            Self::Empty => panic!("state is empty"),
+        }
+    }
+
     fn into_presentation(self) -> Presentation {
         match self {
             Self::Presenting(presentation) => presentation,
@@ -177,6 +246,9 @@ pub enum PresentMode {
 
     /// This is a live presentation so we don't want hot reloading.
     Presentation,
+
+    /// We are running a presentation that's being consumed by `presenterm-export`.
+    Export,
 }
 
 /// An error when loading a presentation.
@@ -199,7 +271,7 @@ pub enum PresentationError {
     Render(#[from] RenderError),
 
     #[error(transparent)]
-    LoadPresentationError(#[from] LoadPresentationError),
+    LoadPresentation(#[from] LoadPresentationError),
 
     #[error("io: {0}")]
     Io(#[from] io::Error),
